//---------------------------------------------------------------
// 	TAP Control 
//---------------------------------------------------------------
module tapcontrol (bsrin, TCK, TDI, TMS, clockBSR, extest, tdiin,
                samplez, shiftDR, TDO, updateDR);

parameter JTAGID=32'h0000a1b3;

input   bsrin;
input   TCK;
input   TDI;
input   TMS;
output  clockBSR;
output  extest;
output  tdiin;
output  samplez;
output  shiftDR;
output  TDO;
output  updateDR;

wire			TDO,TDI,TCK;
reg			tdiin,tmsin;
reg	[3:0]           state;
reg 	[2:0]		ir_reg,ir;
			// 0 - extest
			// 1 - idcode
			// 2 - sampleZ
			// 3 - bypass
			// 4 - sample/preload
			// 5 - mfg
			// 6 - bypass
			// 7 - bypass
reg	[31:0]		id_reg;
reg	[3:0]		mfg_reg;
reg			br_reg;
wire			selIR,selID,selMFG,selBSR,selBR;
wire			extest,samplez,sample;
reg			shiftIR,shiftDR,tdoen,tdout,resetIR;
wire			clockDR,clockBSR,clockIR,clockBR,clockID,clockMFG;
wire			updateIR,updateDR;


reg pu;
initial 
begin
       	pu=1;
    #10	pu=0;
end

assign TDO	=  tdoen ? tdout : 'bz;
assign extest 	= !ir_reg[2] & !ir_reg[1] & !ir_reg[0];
assign selID 	= !ir_reg[2] & !ir_reg[1] &  ir_reg[0];
assign samplez 	= !ir_reg[2] &  ir_reg[1] & !ir_reg[0];
assign selBR 	= (ir_reg[2] &  ir_reg[1])| (ir_reg[1] & ir_reg[0]);
assign selMFG 	=  ir_reg[2] & !ir_reg[1] &  ir_reg[0];
assign sample 	=  ir_reg[2] & !ir_reg[1] & !ir_reg[0];
assign selBSR	=  extest | samplez | sample; 
assign selIR    =  state[3];
assign clockDR	=!(!state[3] &  state[1] & !state[0] & !TCK);
assign clockIR	=!( state[3] &  state[1] & !state[0] & !TCK);
assign clockBR	=!( selBR  & !clockDR); 
assign clockID	=!( selID  & !clockDR); 
assign clockMFG	=!( selMFG & !clockDR); 
assign clockBSR	=!((extest |  samplez | sample) & !clockDR); 
assign updateIR	=  state[3] &  state[2] & !state[1] & state[0] & !TCK;
assign updateDR	= !state[3] &  state[2] & !state[1] & state[0] & !TCK;


// BYPASS
always @(posedge clockBR) 
     if(clockBR)
        br_reg <= shiftDR & tdiin;
// ID
always @(posedge clockID) 
     if(clockID)
        id_reg <= shiftDR ? {tdiin,id_reg[31:1]} : JTAGID;
// MFG
always @(posedge clockMFG) 
     if(clockMFG)
        mfg_reg <= shiftDR ? {tdiin,mfg_reg[3:1]} : 4'bxxxx;
// IR
always @(posedge clockIR) 
     if(clockIR)
	ir <= shiftIR ?  {tdiin,ir[2:1]} : 3'b001;
always @(posedge updateIR or resetIR) 
     if(resetIR)
	ir_reg <= 3'b001;
     else if (updateIR)
	ir_reg <= ir;

always @(negedge TCK or pu ) begin
        resetIR = (state[3] &  state[2] &  state[1] &  state[0])|  pu;
        shiftIR =  state[3] & !state[2] &  state[1] & !state[0] & !pu;
        shiftDR = !state[3] & !state[2] &  state[1] & !state[0] & !pu;
    	tdoen   =             !state[2] &  state[1] & !state[0] & !pu;
	tdout   = selIR ? ir[0] : (selID ? id_reg[0] : (selMFG ? mfg_reg[0] :(selBSR ? bsrin : br_reg)));
        end

parameter          
		jexit2DR     = 0,
                jexit1DR     = 1,
                jshiftDR     = 2,
                jpauseDR     = 3,
                jselectDR    = 4,
                jupdateDR    = 5,
                jcaptureDR   = 6,
                jselectIR    = 7,
                jexit2IR     = 8,
                jexit1IR     = 9,
                jshiftIR     = 10,
                jpauseIR     = 11,
                jidle        = 12,
                jupdateIR    = 13,
                jcaptureIR   = 14,
	        jreset       = 15;

always @(TCK or pu)
	if(pu) begin
		tdiin <= 'b1;
		tmsin <= 'b1;
                state <= jreset;
		end
	else if(TCK==0) begin
	#5 	tdiin <= TDI;
		tmsin <= TMS;
		end

always @(posedge TCK) begin  
    if(TCK==1) begin
         br_reg <= TDI;

        case (state)
        jreset: 
            if (tmsin==0)           state <= jidle;
            else                    state <= jreset;
        jidle :	  
            if (tmsin==0)           state = jidle;
            else                    state = jselectDR;
        jselectDR: 
            if (tmsin==0)           state <= jcaptureDR;
            else                    state <= jselectIR;
        jcaptureDR:
            if (tmsin==0)           state = jshiftDR;
            else                    state = jexit1DR;
        jshiftDR:  
            if (tmsin==0)           state = jshiftDR;
            else                    state = jexit1DR;
        jexit1DR:  
            if (tmsin==0)           state = jpauseDR;
            else                    state = jupdateDR;
        jpauseDR	:  
            if (tmsin==0)           state = jpauseDR;
            else                    state = jexit2DR;
        jexit2DR	:  
            if (tmsin==0)           state = jshiftDR;
            else                    state = jupdateDR;
        jupdateDR : 
            if (tmsin==0)           state = jidle;
            else                    state = jselectDR;
        jselectIR : 
            if (tmsin==0)           state = jcaptureIR;
            else                    state = jreset;
        jcaptureIR : 
            if (tmsin==0)	    state = jshiftIR;
            else                    state = jexit1IR;
        jshiftIR	: 
            if (tmsin==0)           state = jshiftIR;
            else                    state = jexit1IR;
        jexit1IR	:  
            if (tmsin==0)           state = jpauseIR;
            else                    state = jupdateIR;
        jpauseIR	:  
            if (tmsin==0)           state = jpauseIR;
            else                    state = jexit2IR;
        jexit2IR	: 
            if (tmsin==0)           state = jshiftIR;
            else                    state = jupdateIR;
        jupdateIR : 
            if (tmsin==0)           state = jidle;
            else                    state = jselectDR;
        endcase
  end
end 

endmodule  // tapcontrol

module outbsr ( pin, lastc, pineb,  nextc, top, jctl);
input   pin;
input   lastc;
input   pineb;
input [2:0]   jctl;
output  nextc;
output  top;
reg	top;
reg	nextc;
wire    n_2;

//  jctl={updatedr,clockdr,shiftdr};
always @(posedge jctl[1])
	if(jctl[1]) #1 nextc <= (jctl[0] ? lastc : (pin & !pineb));
always @(posedge jctl[2])
	if(jctl[2]) #1 top <= nextc;

endmodule // outbsr

module inbsr (lastc, pin, pineb, nextc, jctl);
input   lastc;
input   pin;
input   pineb;
input [1:0] jctl  ;
output  nextc;
reg nextc;

//   jctl= {clockdr,shiftdr};
always @(posedge jctl[1])
	if(jctl[1]) #1 nextc <= (jctl[0] ? lastc : (pin & !pineb));

endmodule // inbsr

module iobsr (pin, lastc, nextc,n0, pineb, jctl );
inout   pin;
input   lastc;
output  nextc,n0;
input   pineb;
input [4:0] jctl;
//   jctl= {toe,extest,updatedr,clockdr,shiftdr};
wire    top;
wire    n0;
   outbsr i2 (.pin(pin), .lastc(n0), .pineb(pineb), .nextc(nextc), .top(top), .jctl(jctl[2:0]) );
   inbsr  i3 (.pin(pin), .lastc(lastc), .nextc(n0), .pineb(pineb), .jctl(jctl[1:0]) );
assign pin = (jctl[4] & jctl[3]) ? top : 'bz;

endmodule // iobsr

module oebsr (toe, lastc, pin, nextc, jctl);
input   lastc;
input   pin;
output   toe;
input [2:0] jctl  ;
output  nextc;
reg	nextc;
reg	toe;
//   jctl= {updatedr,clockdr,shiftdr};
always @(posedge jctl[1])
	if(jctl[1]) #1 nextc <= (jctl[0] ? lastc : pin );
always @(posedge jctl[2])
	if(jctl[2]) #1 toe <= nextc;

endmodule // oebsr
